import ubersmith
import ubersmith.api


def setup_module():
    ubersmith.init(**{
        'base_url': '',
        'username': '',
        'password': '',
    })


def teardown_module():
    ubersmith.api._DEFAULT_REQUEST_HANDLER = None



"""Utility functions used throughout ubersmith library."""
import inspect
try:
    from collections.abc import MutableSequence, MutableMapping
except ImportError:  # pragma: no cover
    from collections import MutableSequence, MutableMapping
try:
    from urllib import parse as urlparse
except ImportError:  # pragma: no cover
    import urlparse
try:
    from urllib.parse import urlencode
except ImportError:  # pragma: no cover
    from urllib import urlencode

from six import string_types, text_type

__all__ = [
    'append_qs',
    'urlencode_unicode',
    'to_nested_php_args',
    'prepend_base',
    'isdict',
    'islist',
    'isstr',
    'get_filename'
]


def append_qs(url, query_string):
    """Append query_string values to an existing URL and return it as a string.

    query_string can be:
        * an encoded string: 'test3=val1&test3=val2'
        * a dict of strings: {'test3': 'val'}
        * a dict of lists of strings: {'test3': ['val1', 'val2']}
        * a list of tuples: [('test3', 'val1'), ('test3', 'val2')]

    """
    parsed_url = urlparse.urlsplit(url)
    parsed_qs = urlparse.parse_qsl(parsed_url.query, True)

    if isstr(query_string):
        parsed_qs += urlparse.parse_qsl(query_string)
    elif isdict(query_string):
        for item in query_string.items():
            if islist(item[1]):
                for val in item[1]:
                    parsed_qs.append((item[0], val))
            else:
                parsed_qs.append(item)
    elif islist(query_string):
        parsed_qs += query_string
    else:
        raise TypeError('Unexpected query_string type')

    return urlparse.urlunsplit((
        parsed_url.scheme,
        parsed_url.netloc,
        parsed_url.path,
        urlencode_unicode(parsed_qs),
        parsed_url.fragment,
    ))


def urlencode_unicode(data, doseq=0):
    """urllib.urlencode can't handle unicode, this is a hack to fix it."""
    data_iter = None
    if isdict(data):
        data_iter = data.items()
    elif islist(data):
        data_iter = data

    if data_iter:
        for i, (key, value) in enumerate(data_iter):
            if isinstance(value, text_type):
                # try to convert to str
                try:
                    safe_val = str(value)
                except UnicodeEncodeError:
                    # try to encode as utf-8
                    # if an exception is raised here then idk what to do
                    safe_val = value.encode('utf-8')
                finally:
                    if isdict(data):
                        data[key] = safe_val
                    else:
                        data[i] = (key, safe_val)

    return urlencode(data, doseq=doseq)


def _is_leaf(value):
    """Return if a value is a leaf to an existing object"""
    return not (isdict(value) or islist(value))


def to_nested_php_args(data, prefix_key=None):
    """
    This function will take either a dict or list and will recursively loop
    through the values converting it into a format similar to a PHP array which
    Ubersmith requires for the info portion of the API's order.create method.
    """
    is_root = prefix_key is None
    prefix_key = prefix_key if prefix_key else ''

    if islist(data):
        data_iter = data if is_root else enumerate(data)
        new_data = [] if is_root else {}
    elif isdict(data):
        data_iter = data.items()
        new_data = {}
    else:
        raise TypeError('expected dict or list, got {0}'.format(type(data)))

    if islist(new_data):
        def data_set(k, v):
            new_data.append((k, v))
        def data_update(d):
            for k, v in d.items():
                new_data.append((k, v))
    else:
        def data_set(k, v):
            new_data[k] = v
        data_update = new_data.update

    for key, value in data_iter:
        end_key = prefix_key + (str(key) if is_root else '[{0}]'.format(key))
        if _is_leaf(value):
            data_set(end_key, value)
        else:
            nested_args = to_nested_php_args(value, end_key)
            data_update(nested_args)

    return new_data


def prepend_base(base):
    """Return a callable that will prepend the base of a method string."""
    return lambda call: '.'.join((base, call))


def isdict(value):
    """Return true if the value behaves like a dict, false if not."""
    return isinstance(value, MutableMapping)


def islist(value):
    """Return true if the value behaves like a list, false if not."""
    return isinstance(value, MutableSequence)


def isstr(value):
    """Return true if the value behaves like a string, false if not."""
    return isinstance(value, string_types)


def get_filename(disposition):
    """Parse Content-Disposition header to pull out the filename bit.

    See: http://tools.ietf.org/html/rfc2616#section-19.5.1

    """
    if disposition:
        params = [param.strip() for param in disposition.split(';')[1:]]
        for param in params:
            if '=' in param:
                name, value = param.split('=', 1)
                if name == 'filename':
                    return value.strip('"')



"""Order call classes.

These classes implement any response cleaning and validation needed.  If a
call class isn't defined for a given method then one is created using
ubersmith.calls.BaseCall.

"""

from ubersmith.calls import BaseCall
from ubersmith.clean import clean
from ubersmith.utils import prepend_base

__all__ = [
    'GetCall',
    'ListCall',
    'QueueListCall',
]

_ = prepend_base(__name__.split('.')[-1])


class GetCall(BaseCall):
    method = _('get')
    required_fields = [('order_id', 'hash')]
    cleaner = clean(dict, values={
        'order_id': 'int',
        'order_status': 'int',
        'client_id': 'int',
        'order_form_id': 'int',
        'order_queue_id': 'int',
        'opportunity_id': 'int',
        'total': 'decimal',
        'activity': 'timestamp',
        'ts': 'timestamp',
        'progress': clean(dict, keys='int', values=clean(dict, values={
            'ts': 'timestamp',
        })),
    })


class ListCall(BaseCall):
    method = _('list')
    cleaner = clean(dict, keys='int', values=clean(dict, values={
        'order_id': 'int',
        'priority': 'int',
        'order_status': 'int',
        'client_id': 'int',
        'order_form_id': 'int',
        'order_queue_id': 'int',
        'opportunity_id': 'int',
        'total': 'decimal',
        'activity': 'timestamp',
        'ts': 'timestamp',
    }))


class QueueListCall(BaseCall):
    method = _('queue_list')
    required_fields = ['brand_id']
    cleaner = clean(dict, keys='int', values=clean(dict, values={
        'steps': clean(dict, keys='int', values=clean(dict, values={
            'count': 'int',
        })),
    }))



import datetime
from decimal import Decimal
import json

import pytest
from six import text_type

import ubersmith
import ubersmith.api
import ubersmith.client


def setup_module():
    ubersmith.init(**{
        'base_url': '',
        'username': '',
        'password': '',
    })


def teardown_module():
    ubersmith.api._DEFAULT_REQUEST_HANDLER = None


def test_get(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {
            u'clientid': u'50',
            u'access': u'a:1:{s:4:"cbms";s:6:"client";}',
            u'active': u'1',
            u'balance': u'187559.67',
            u'class_id': u'1',
            u'created': u'1236056400',
            u'discount': u'100.00',
            u'latest_inv': u'1278014814',
            u'password_changed': u'1304603745',
            u'priority': u'1',
            u'tier_commission_rate': u'0.00',
            u'tier_commission': u'0.00',
            u'commission_rate': u'0.00',
            u'commission': u'0.00',
        },
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    expected = {
        u'clientid': 50,
        u'access': {b'cbms': b'client'},
        u'active': True,
        u'balance': Decimal('187559.67'),
        u'class_id': 1,
        u'created': datetime.datetime.fromtimestamp(float('1236056400')),
        u'discount': Decimal('100.00'),
        u'latest_inv': datetime.datetime.fromtimestamp(float('1278014814')),
        u'password_changed': datetime.datetime.fromtimestamp(float('1304603745')),
        u'priority': 1,
        u'tier_commission_rate': Decimal('0.00'),
        u'tier_commission': Decimal('0.00'),
        u'commission_rate': Decimal('0.00'),
        u'commission': Decimal('0.00'),
    }
    assert dict(ubersmith.client.get(client_id=50)) == expected


def test_update(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {},
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    expected = {}
    result = ubersmith.client.get(client_id=50, email="bob@example.com")
    assert dict(result) == expected


def test_invoice_list(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {
            "60": {
                u'invid': u'60',
                u'clientid': u'50',
                u'amount': u'50.00',
                u'amount_unpaid': u'0.00',
                u'date': u'1272400333',
                u'datepaid': u'1272400333',
                u'due': u'1272400333',
            },
        },
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    expected = {
        60: {
            u'invid': 60,
            u'clientid': 50,
            u'amount': u'50.00',
            u'amount_unpaid': u'0.00',
            u'date': datetime.datetime.fromtimestamp(float("1272400333")),
            u'datepaid': datetime.datetime.fromtimestamp(float("1272400333")),
            u'due': datetime.datetime.fromtimestamp(float("1272400333")),
        }
    }
    assert dict(ubersmith.client.invoice_list(client_id=50)) == expected


def test_invoice_count(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": "42",
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    assert int(ubersmith.client.invoice_count(client_id=50)) == 42


def test_invoice_get(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {
            u'invid': u'60',
            u'clientid': u'50',
            u'date': u'1272400333',
            u'datepaid': u'1272400333',
            u'due': u'1272400333',
            u'overdue': u'1272400333',
        },
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    expected = {
        u'invid': 60,
        u'clientid': 50,
        u'date': datetime.datetime.fromtimestamp(float("1272400333")),
        u'datepaid': datetime.datetime.fromtimestamp(float("1272400333")),
        u'due': datetime.datetime.fromtimestamp(float("1272400333")),
        u'overdue': datetime.datetime.fromtimestamp(float("1272400333")),
    }
    assert dict(ubersmith.client.invoice_get(invoice_id=60)) == expected


def test_invoice_get_pdf(response):
    response.headers = {
        'content-type': 'application/pdf',
        'content-disposition': 'inline; filename=Invoice-60.pdf'
    }
    response.text = u"Some PDF data."
    response.content = str(response.text)
    uberfile = ubersmith.client.invoice_get(invoice_id=60, format="pdf")
    assert uberfile.type == 'application/pdf'
    assert uberfile.filename == "Invoice-60.pdf"
    assert str(uberfile.data) == response.text


def test_invoice_get_pdf_without_disposition(response):
    response.headers = {
        'content-type': 'application/pdf',
    }
    response.text = u"Some PDF data."
    response.content = str(response.text)
    uberfile = ubersmith.client.invoice_get(invoice_id=60, format="pdf")
    assert uberfile.type == 'application/pdf'
    assert uberfile.filename is None
    assert str(uberfile.data) == response.text



"""Sales call functions.

These are light weight call functions that basically just wrap call classes
under ubersmith.calls.  If a call function doesn't exist it will be generated
by generate_generic_calls which searches for a call class and if one isn't
found one is created using ubersmith.calls.BaseCall.

"""

from ubersmith.calls import generate_generic_calls

__all__ = []


generate_generic_calls(__name__.split('.')[-1], globals())



"""Device call classes.

These classes implement any response cleaning and validation needed.  If a
call class isn't defined for a given method then one is created using
ubersmith.calls.BaseCall.

"""

from ubersmith.calls import BaseCall
from ubersmith.clean import clean
from ubersmith.utils import prepend_base

__all__ = [
    'GetCall',
    'ListCall',
    # 'ModuleGraphCall',
    'MonitorListCall',
]

_ = prepend_base(__name__.split('.')[-1])


_DEVICE_CLEANER = clean(dict, values={
    'active': 'int',
    'cage_id': 'int',
    'clientid': 'int',
    'dev': 'int',
    'devtype_group_id': 'int',
    'disabled': 'int',
    'down': 'int',
    'fac_id': 'int',
    'owner': 'int',
    'parent': 'int',
    'rack_id': 'int',
    'row_id': 'int',
    'total': 'int',
    'type_id': 'int',
    'up': 'int',
    'warn': 'int',
    'zone_id': 'int',
    'depth': float,
    'height': float,
    'width': float,
})


class GetCall(BaseCall):
    method = _('get')
    required_fields = ['device_id']
    cleaner = _DEVICE_CLEANER


class ListCall(BaseCall):
    method = _('list')
    cleaner = clean(dict, keys='int', values=_DEVICE_CLEANER)


class IpAssignmentListCall(BaseCall):
    method = _('ip_assignment_list')
    cleaner = clean(dict, keys='int', values=clean(dict, values={
        'addr_type': 'int',
        'assign_id': 'int',
        'client_id': 'int',
        'created_ts': 'timestamp',
        'device_id': 'int',
        'group_client_id': 'int',
        'group_id': 'int',
        'num_ips': 'int',
        'portable': bool,
        'service_id': 'int',
        'updated_ts': 'timestamp',
        'vlan_num': 'int',
        'vlan_range_id': 'int',
        'vlan_type_id': 'int',
    }))


class ModuleGraphCall(BaseCall):
    method = _('module_graph')


class MonitorListCall(BaseCall):
    method = _('monitor_list')
    required_fields = ['protocol']
    cleaner = clean(dict, keys='int', values=clean(dict, values={
        'dev': 'int',
        'script_id': 'int',
        'last_change': 'timestamp',
        'last_notified': 'timestamp',
        'last_poll': 'timestamp',
    }))



from mock import Mock
import pytest

import ubersmith.api


@pytest.fixture
def requests(monkeypatch):
    requests = Mock()
    monkeypatch.setattr(ubersmith.api, 'requests', requests)
    return requests


@pytest.fixture
def response(requests):
    response = Mock()
    response.headers = {
        'content-type': 'application/json',
    }
    response.text = u""
    requests.post.return_value = response
    return response



"""Client call classes.

These classes implement any response cleaning and validation needed.  If a
call class isn't defined for a given method then one is created using
ubersmith.calls.BaseCall.

"""

from collections import namedtuple

from ubersmith.calls import BaseCall
from ubersmith.clean import clean
from ubersmith.utils import prepend_base, get_filename

__all__ = [
    'GetCall',
    'ListCall',
    'PaymentMethodListCall',
    'InvoiceCountCall',
    'InvoicePaymentsCall',
    'InvoiceGet',
    'InvoiceList',
    'CreditListCall',
]

_ = prepend_base(__name__.split('.')[-1])


_CLIENT_CLEANER = clean(dict, values={
    'active': bool,
    'clientid': 'int',
    'class_id': 'int',
    'priority': 'int',
    'balance': 'decimal',
    'commission': 'decimal',
    'commission_rate': 'decimal',
    'discount': 'decimal',
    'tier_commission': 'decimal',
    'tier_commission_rate': 'decimal',
    'created': 'timestamp',
    'latest_inv': 'timestamp',
    'password_changed': 'timestamp',
    'access': 'php',
})


class GetCall(BaseCall):
    method = _('get')
    required_fields = [('client_id', 'user_login', 'email')]
    cleaner = _CLIENT_CLEANER


class ListCall(BaseCall):
    method = _('list')
    cleaner = clean(dict, keys='int', values=_CLIENT_CLEANER)


class PaymentMethodListCall(BaseCall):
    method = _('payment_method_list')
    cleaner = clean(dict, keys='int')


class InvoiceCountCall(BaseCall):
    method = _('invoice_count')
    required_fields = ['client_id']
    cleaner = int


class InvoicePaymentsCall(BaseCall):
    method = _('invoice_payments')
    required_fields = ['invoice_id']
    cleaner = clean(dict, keys='int', values=clean(dict, values={
        'time': 'timestamp',
    }))


class InvoiceGet(BaseCall):
    method = _('invoice_get')
    required_fields = ['invoice_id']
    cleaner = clean(dict, values={
        'clientid': 'int',
        'invid': 'int',
        'date': 'timestamp',
        'datepaid': 'timestamp',
        'due': 'timestamp',
        'overdue': 'timestamp',
    })


class InvoiceList(BaseCall):
    method = _('invoice_list')
    cleaner = clean(dict, keys='int', values=clean(dict, values={
        'clientid': 'int',
        'invid': 'int',
        'date': 'timestamp',
        'datepaid': 'timestamp',
        'due': 'timestamp',
    }))


class CreditListCall(BaseCall):
    method = _('credit_list')
    required_fields = ['client_id']
    cleaner = clean(dict, keys='int', values=clean(dict, values={
        'clientid': 'int',
        'active': 'int',
        'credit_id': 'int',
        'order_id': 'int',
        'date': 'timestamp',
    }))



import datetime
from decimal import Decimal
import json

from six import text_type

import ubersmith
import ubersmith.api


def setup_module():
    ubersmith.init(**{
        'base_url': '',
        'username': '',
        'password': '',
    })


def teardown_module():
    ubersmith.api._DEFAULT_REQUEST_HANDLER = None


def test_order_get(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {
            "order_id": "60",
            "order_status": "3",
            "client_id": "50",
            "order_form_id": "12",
            "order_queue_id": "4",
            "opportunity_id": "243",
            "total": "1,234.5",
            "activity": "1272400333",
            "ts": "1272400333",
            "progress": {
                "1": {
                    "ts": "1272400333",
                },
                "2": {
                    "ts": "1272400333",
                },
                "3": {
                    "ts": "1272400333",
                },
            },
        },
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    expected = {
        u'order_id': 60,
        u'order_status': 3,
        u'client_id': 50,
        u'order_form_id': 12,
        u'order_queue_id': 4,
        u'opportunity_id': 243,
        u'total': Decimal('1234.5'),
        u'activity': datetime.datetime.fromtimestamp(float('1272400333')),
        u'ts': datetime.datetime.fromtimestamp(float('1272400333')),
        u'progress': {
            1: {
                "ts": datetime.datetime.fromtimestamp(float("1272400333")),
            },
            2: {
                "ts": datetime.datetime.fromtimestamp(float("1272400333")),
            },
            3: {
                "ts": datetime.datetime.fromtimestamp(float("1272400333")),
            },
        },
    }
    assert dict(ubersmith.order.get(order_id=60)) == expected


def test_order_list(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {
            "60": {
                "order_id": "60",
                "client_id": "50",
                "activity": "1272400333",
                "ts": "1272400333",
                "total": "33.22",
            },
        },
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    expected = {
        60: {
            u'client_id': 50,
            u'activity': datetime.datetime.fromtimestamp(float("1272400333")),
            u'ts': datetime.datetime.fromtimestamp(float("1272400333")),
            u'total': Decimal('33.22'),
            u'order_id': 60,
        }
    }
    assert dict(ubersmith.order.list(client_id=50)) == expected


def test_order_queue_list(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {
            u'1': {
                u'name': u'Orders',
                u'steps': {
                    u'1': {u'count': u'10', u'name': u'foo'},
                    u'2': {u'count': u'20', u'name': u'bar'},
                    u'3': {u'count': u'30', u'name': u'baz'},
                },
            },
        },
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    expected = {
        1: {
            u'name': u'Orders',
            u'steps': {
                1: {u'count': 10, u'name': u'foo'},
                2: {u'count': 20, u'name': u'bar'},
                3: {u'count': 30, u'name': u'baz'},
            },
        },
    }
    assert dict(ubersmith.order.queue_list(brand_id=1)) == expected



"""Uber call classes.

These classes implement any response cleaning and validation needed.  If a
call class isn't defined for a given method then one is created using
ubersmith.calls.BaseCall.

"""

from ubersmith.calls import BaseCall
from ubersmith.clean import clean
from ubersmith.utils import prepend_base

__all__ = [
    'ApiExportCall',
    'CheckLoginCall',
    'ClientWelcomeStatsCall',
    'MethodGetCall',
    'DocumentationCall',
]

_ = prepend_base(__name__.split('.')[-1])


class ApiExportCall(BaseCall):
    method = _('api_export')
    required_fields = ['table']


class CheckLoginCall(BaseCall):
    method = _('check_login')
    cleaner = clean(dict, values={
        'password_expired': 'int',
        'last_login': 'timestamp',
        'password_changed': 'timestamp',
    })


class ClientWelcomeStatsCall(BaseCall):
    method = _('client_welcome_stats')
    required_fields = ['client_id']
    cleaner = clean(dict, values={
        'client_activity_type': 'int',
        'closed_count': 'int',
        'inv_count': 'int',
        'pack_count': 'int',
        'ticket': 'int',
        'type': 'int',
        'client_activity': 'timestamp',
        'next_inv': 'date',
    })


class MethodGetCall(BaseCall):
    method = _('method_get')


class DocumentationCall(BaseCall):
    method = _('documentation')



"""Lower level API, configuration, and HTTP stuff."""
import six
import time
from ubersmith.compat import total_ordering, file_type

import requests

from ubersmith.exceptions import (
    RequestError,
    ResponseError,
    UpdatingTokenResponse,
    MaintenanceResponse,
)
from ubersmith.utils import (
    append_qs,
    to_nested_php_args,
    get_filename,
)

__all__ = [
    'METHODS',
    'RequestHandler',
    'get_default_request_handler',
    'set_default_request_handler',
]

_DEFAULT_REQUEST_HANDLER = None

"""A dict of all methods returned by uber.method_list()"""
METHODS = {
    u'client.ach_add': u'Add a New Bank Account',
    u'client.ach_delete': u'Delete a Bank Account',
    u'client.ach_update': u'Update a Bank Account',
    u'client.add': u'Add a New Client',
    u'client.avatar_get': u"Retrieve a Client Avatar",
    u'client.avatar_set': u"Set a Client Avatar",
    u'client.cc_add': u'Add a New Credit Card',
    u'client.cc_delete': u'Delete a Credit Card',
    u'client.cc_info': u"List a Client's Credit Card Details",
    u'client.cc_update': u'Update a Credit Card',
    u'client.comment_list': u"List a Client's Comments",
    u'client.contact_add': u'Add a New Contact',
    u'client.contact_delete': u'Delete a Contact',
    u'client.contact_get': u'Get Contact Details',
    u'client.contact_list': u"List a Client's Contacts",
    u'client.contact_metadata_get': u"Get a Contact's Metadata",
    u'client.contact_metadata_single': u"Get a Contact's Metadata Value",
    u'client.contact_update': u'Update a Contact',
    u'client.count': u'Count Active Clients',
    u'client.credit_add': u'Add an Account Credit',
    u'client.credit_apply': u"Apply a Credit to an Invoice",
    u'client.credit_comment_list': u"List a Credit's Comments",
    u'client.credit_deactivate': u'Deactivate an Account Credit',
    u'client.credit_list': u"List a Client's Credits",
    u'client.deactivate': u'Deactivate a Client',
    u'client.domain_add': u'Add a Domain',
    u'client.domain_list': u"List a Client's Domains",
    u'client.domain_lookup': u'Look Up a Domain',
    u'client.domain_register': u'Register a Domain',
    u'client.domain_transfer': u'Transfer a Domain',
    u'client.get': u'Get Client Details',
    u'client.invoice_charge': u'Charge an Invoice',
    u'client.invoice_count': u'Count Invoices',
    u'client.invoice_disregard': u'Disregard an Invoice',
    u'client.invoice_generate': u'Generate an Invoice',
    u'client.invoice_get': u'Get an Invoice',
    u'client.invoice_list': u"List a Client's Invoices",
    u'client.invoice_payments': u"List an Invoice's Payments",
    u'client.invoice_post_gw_payment': u'Record a Payment',
    u'client.latest_client': u'Get the Latest Client',
    u'client.list': u'List Clients',
    u'client.lookup': u'Look Up a Client',
    u'client.metadata_get': u"Get a Client's Metadata",
    u'client.metadata_single': u"Get a Client's Metadata Value",
    u'client.payment_method_list': u"List a Client's Payment Methods",
    u'client.payment_refund': u'Refund a payment.',
    u'client.reactivate': u'Reactivate a Client',
    u'client.renewal_list': u'List Services for Renewal',
    u'client.send_welcome': u'Send a Welcome Letter',
    u'client.service_add': u'Add a New Service',
    u'client.service_comment_list': u"List a Service's Comments",
    u'client.service_deactivate': u'Deactivate a Service',
    u'client.service_get': u'Get a Service',
    u'client.service_list': u"List a Client's Services",
    u'client.service_metadata_get': u"Get a Service's Metadata",
    u'client.service_metadata_single': u"Get a Service's Metadata Value",
    u'client.service_module_call': u'Call a Service Module Function',
    u'client.service_prorate': u'Prorate a Service',
    u'client.service_update': u'Update a Service',
    u'client.set_login': u"Set a Client's Login",
    u'client.tax_exemption_add': u"Add a new Tax Exemption",
    u'client.tax_exemption_get': u"Get a Client's Tax Exemption",
    u'client.tax_exemption_list': u"List a Client's Tax Exemptions",
    u'client.tax_exemption_update': u"Update a Client's Tax Exemption",
    u'client.update': u'Update a Client',
    u'device.add': u'Add a New Device',
    u'device.comment_list': u"List a Device's Comments",
    u'device.cpanel_add': u'Add a cPanel Account',
    u'device.delete': u'Delete a Device',
    u'device.event_list': u'List Device Events',
    u'device.facility_list': u'List Device Facilities',
    u'device.get': u'Get a Device',
    u'device.hostname_get': u'Get a Device Hostname',
    u'device.ip_assign': u'Assign an IP to a Device',
    u'device.ip_assignment_add': u'Create a New IP Assignment',
    u'device.ip_assignment_delete': u'Delete a Device IP Assignment',
    u'device.ip_assignment_list': u'List Device IP Assignments',
    u'device.ip_assignment_update': u'Update a Device IP Assignment',
    u'device.ip_block_list': u"List IP Blocks",
    u'device.ip_get_available': u'List Available IP Addresses',
    u'device.ip_get_unassigned': u'Get Unassigned IP Addresses',
    u'device.ip_group_add': u'Add a Device IP Group',
    u'device.ip_group_delete': u'Delete a Device IP Group',
    u'device.ip_group_list': u'List a Device IP Group',
    u'device.ip_group_update': u'Update a Device IP Group',
    u'device.ip_lookup': u'Look Up a Device IP',
    u'device.ip_pool_list': u"List IP Pools",
    u'device.ip_unassign': u'Unassign a Device IP',
    u'device.list': u'List Devices',
    u'device.module_call': u'Call a Device Module Function',
    u'device.module_call_aggregate': u'Call an Aggregate Device Module Function',
    u'device.module_graph': u'Generate Device Module Graph',
    u'device.monitor_add': u'Add a New Device Monitor',
    u'device.monitor_delete': u'Delete a Device Monitor',
    u'device.monitor_disable': u'Disable a Device Monitor',
    u'device.monitor_enable': u'Enable a Device Monitor',
    u'device.monitor_list': u'List Device Monitors',
    u'device.monitor_update': u'Update a Device Monitor',
    u'device.reboot': u"Set a Device's Power State",
    u'device.reboot_graph': u'Get a Reboot Graph',
    u'device.tag': u'Tag a Device',
    u'device.type_list': u'List Device Types',
    u'device.untag': u'Untag a Device',
    u'device.update': u'Update a Device',
    u'device.vlan_get_available': u'List Available VLANs',
    u'order.cancel': u'Cancel an Order',
    u'order.client_respond': u'Post a Client/Lead Order Response',
    u'order.coupon_get': u'Get Order Coupon Details',
    u'order.create': u'Create a New Order',
    u'order.get': u'Get Order Details',
    u'order.list': u'List Orders',
    u'order.process': u'Process an Order',
    u'order.queue_list': u'List Order Queues',
    u'order.respond': u'Post an Order Response',
    u'order.submit': u'Submit An Order',
    u'order.update': u'Update an Order',
    u'sales.opportunity_add': u'Add an Opportunity',
    u'sales.opportunity_list': u'List Opportunities',
    u'sales.opportunity_stage_list': u'List Opportunity Stages',
    u'sales.opportunity_status_list': u'List Opportunity Statuses',
    u'sales.opportunity_type_list': u'List Opportunity Types',
    u'sales.opportunity_update': u'Update an Opportunity',
    u'support.department_get': u'Get Ticket Departments',
    u'support.department_list': u'List Ticket Departments',
    u'support.ticket_count': u'Count Support Tickets',
    u'support.ticket_get': u'Get Support Ticket Details',
    u'support.ticket_list': u'Get a List of Tickets',
    u'support.ticket_merge': u'Merge Tickets',
    u'support.ticket_post_client_response': u'Post a Client Response to a Ticket',
    u'support.ticket_post_list': u'Get all Posts for a Ticket',
    u'support.ticket_post_staff_response': u'Post a Staff Response to a Ticket',
    u'support.ticket_submit': u'Submit a New Ticket',
    u'support.ticket_submit_outgoing': u'Create a New Outgoing Ticket',
    u'support.ticket_type_list': u"Get a List of Ticket Types",
    u'support.ticket_update': u'Update a Ticket',
    u'uber.admin_avatar_get': u"Retrieve an Admin Avatar",
    u'uber.admin_avatar_set': u"Set an Admin Avatar",
    u'uber.admin_get': u"User Information",
    u'uber.admin_list': u"List User Logins",
    u'uber.api_export': u'Export Data',
    u'uber.attachment_get': u'Get an attachment',
    u'uber.attachment_list': u'List Attachments',
    u'uber.check_login': u'Verify a login and password',
    u'uber.client_permission_list': u"List available permissions",
    u'uber.client_welcome_stats': u'Display Client Statistics',
    u'uber.comment_add': u'Add Comment',
    u'uber.comment_delete': u'Delete Comment',
    u'uber.comment_get': u'Get Comments',
    u'uber.comment_list': u'List Comments',
    u'uber.comment_update': u'Update Comment',
    u'uber.documentation': u'Download API Documentation',
    u'uber.event_list': u'Access the Event Log',
    u'uber.file_add': u"Add a file",
    u'uber.file_delete': u"Delete a file",
    u'uber.file_get': u"Get a File",
    u'uber.file_list': u"Get a List of Files",
    u'uber.file_update': u"Update a file",
    u'uber.forgot_pass': u'Send a Password Reminder',
    u'uber.login_list': u'List User Logins',
    u'uber.mail_get': u'Get an Email From the Log',
    u'uber.mail_list': u'Access the Mail Log',
    u'uber.message_list': u'List Message Board Messages',
    u'uber.metadata_bulk_get': u'Bulk Get Metadata Values',
    u'uber.metadata_get': u'Get Metadata Values',
    u'uber.method_get': u'Get API Method Details',
    u'uber.method_list': u'List Available API Methods',
    u'uber.quick_stats': u'Get Quick System Stats',
    u'uber.quick_stats_detail': u'Get Detailed System Stats',
    u'uber.service_plan_get': u'Get Service Plan Details',
    u'uber.service_plan_list': u'List Service Plans',
    u'uber.tax_exemption_type_get': u"Get a Tax Exemption Type",
    u'uber.tax_exemption_type_list': u"List Tax Exemption Types",
    u'uber.user_exists': u'Check whether a Client Exists',
    u'uber.username_exists': u'Check Whether a Username Exists',
}


class _ProxyModule(object):
    def __init__(self, handler, module):
        self.handler = handler
        self.module = module

    def __getattr__(self, name):
        """Return the call with request_handler prefilled."""
        call_func = getattr(self.module, name)
        if callable(call_func):
            call_p = call_func.handler(self.handler)
            # store partial on proxy so it doesn't have to be created again
            setattr(self, name, call_p)
            return call_p
        raise AttributeError("'{0}' object has no attribute '{1}'".format(
            type(self).__name__, name))


class RequestHandler(object):
    """Handles HTTP requests and authentication."""

    def __init__(self, base_url, username=None, password=None, verify=True):
        """Initialize HTTP request handler with optional authentication.

            base_url: URL to send API requests
            username: Username for API access
            password: Password for API access
            verify: Verify HTTPS certificate

        """
        self.base_url = base_url
        self.username = username
        self.password = password
        self.verify = verify

    def process_request(self, method, data=None):
        """Process request over HTTP to ubersmith instance.

            method: Ubersmith API method string
            data: dict of method arguments

        """
        # make sure requested method is valid
        self._validate_request_method(method)

        # attempt the request three times
        attempts = range(3)
        for i in attempts:
            response = self._send_request(method, data)

            # handle case where ubersmith is 'updating token'
            # see: https://github.com/jasonkeene/python-ubersmith/issues/1
            if all([
                response.headers.get('content-type') == 'text/html',
                'Updating Token' in response.content,
            ]):
                if i != attempts[-1]:
                    # wait 2 secs before retrying request
                    time.sleep(2)
                    continue
                else:
                    raise UpdatingTokenResponse
            break

        resp = BaseResponse(response)

        # test for error in json response
        if response.headers.get('content-type') == 'application/json':
            if not resp.json.get('status'):
                if all([
                    resp.json.get('error_code') == 1,
                    resp.json.get('error_message') == u"We are currently "
                        "undergoing maintenance, please check back shortly.",
                ]):
                    raise MaintenanceResponse(response=resp.json)
                else:
                    raise ResponseError(response=resp.json)
        return resp

    def _send_request(self, method, data):
        url = append_qs(self.base_url, {'method': method})
        data, files, headers = self._encode_data(data)
        return requests.post(url, data=data, files=files, headers=headers,
                             auth=(self.username, self.password),
                             verify=self.verify)

    @staticmethod
    def _validate_request_method(method):
        """Make sure requested method is valid."""
        if method not in METHODS:
            raise RequestError("Requested method is not valid.")

    @staticmethod
    def _encode_data(data):
        """URL encode data."""
        data = data if data is not None else {}
        data = to_nested_php_args(data)
        files = dict([
            (key, value) for key, value in
            data.items() if isinstance(value, file_type)])
        for fname in files:
            del data[fname]
        return data, files or None, None

    def __getattr__(self, name):
        """If attribute accessed is a call module, return a proxy."""
        if name in set(m.split('.')[0] for m in METHODS):
            module_name = 'ubersmith.{0}'.format(name)
            module = __import__(module_name, fromlist=[''])
            proxy = _ProxyModule(self, module)
            # store proxy on handler so it doesn't have to be created again
            setattr(self, name, proxy)
            return proxy
        raise AttributeError("'{0}' object has no attribute '{1}'".format(
            type(self).__name__, name))


class BaseResponse(object):
    """Wraps response object and emulates different types."""
    def __init__(self, response):
        self.response = response  # requests' response object

    @classmethod
    def from_cleaned(cls, response, cleaned):
        resp = cls(response.response)
        resp.cleaned = cleaned
        return resp

    @property
    def json(self):
        return self.response.json()

    @property
    def data(self):
        if hasattr(self, "cleaned"):
            return self.cleaned
        else:
            return self.json['data']

    @property
    def type(self):
        return self.response.headers.get('content-type')

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return repr(self.data)

    def __nonzero__(self):
        return bool(self.data)

    def __json__(self):
        """This method returns the JSON-serializable representation of the
        Response. To utilize this, create a JSONEncoder which calls the
        __json__ methods of supporting objects. e.g.::

            import json
            class MyJSONEncoder(json.JSONEncoder):
                def default(self, o):
                    if hasattr(obj, '__json__') and callable(obj.__json__):
                        return obj.__json__()
                    else:
                        return super(MyJSONEncoder, self).default(o)

            json.dumps(my_response, cls=MyJSONEncoder)
        """
        return self.data


@total_ordering
class DictResponse(BaseResponse):
    __marker = object()

    def keys(self):
        return self.data.keys()

    def iterkeys(self):
        return six.iterkeys(self.data)

    def values(self):
        return self.data.values()

    def itervalues(self):
        return six.itervalues(self.data)

    def items(self):
        return self.data.items()

    def iteritems(self):
        return six.iteritems(self.data)

    def get(self, key, default=None):
        return self.data.get(key, default)

    def update(self, d):
        self.data.update(d)

    def setdefault(self, key, value):
        self.data.setdefault(key, value)

    def pop(self, key, default=__marker):
        if default is self.__marker:
            return self.data.pop(key)
        else:
            return self.data.pop(key, default)

    def popitem(self):
        return self.data.popitem()

    def clear(self):
        self.data.clear()

    def __setitem__(self, key, value):
        self.data[key] = value

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, key):
        return self.data[key]

    def __len__(self):
        return len(self.data)

    def __eq__(self, other):
        return self.data == other

    def __lt__(self, other):
        return self.data < other

    def __contains__(self, item):
        return item in self.data


@total_ordering
class IntResponse(BaseResponse):
    @property
    def numerator(self):
        return self.data

    @property
    def denominator(self):
        return 1

    @property
    def real(self):
        return self.data

    @property
    def imag(self):
        return 0

    def bit_length(self):
        if hasattr(self.data, 'bit_length'):
            return self.data.bit_length()
        else:
            return len(bin(abs(self.data))) - 2

    def conjugate(self):
        return self.data

    def __int__(self):
        return self.data
    __index__ = __long__ = __trunc__ = __int__

    def __float__(self):
        return float(self.data)

    def __oct__(self):
        return oct(self.data)

    def __hex__(self):
        return hex(self.data)

    def __eq__(self, other):
        return self.data == other

    def __lt__(self, other):
        return self.data < other

    def __add__(self, other):
        return int(self) + other
    __radd__ = __add__

    def __sub__(self, other):
        return int(self) - other

    def __rsub__(self, other):
        return other - int(self)

    def __mul__(self, other):
        return int(self) * other
    __rmul__ = __mul__

    def __div__(self, other):
        return int(self) / other

    def __rdiv__(self, other):
        return other / int(self)

    def __floordiv__(self, other):
        return int(self) // other

    def __rfloordiv__(self, other):
        return other // int(self)

    def __truediv__(self, other):
        return float(self) / other

    def __rtruediv__(self, other):
        return other / float(self)

    def __mod__(self, other):
        return int(self) % other

    def __rmod__(self, other):
        return other % int(self)

    def __pow__(self, other):
        return int(self) ** other

    def __rpow__(self, other):
        return other ** int(self)

    def __abs__(self):
        return abs(self.data)

    def __neg__(self):
        return -self.data

    def __pos__(self):
        return self.data

    def __divmod__(self, other):
        return self // other, self % other

    def __rdivmod__(self, other):
        return other // self, other % self

    def __and__(self, other):
        return self.data & other
    __rand__ = __and__

    def __or__(self, other):
        return self.data | other
    __ror__ = __or__

    def __xor__(self, other):
        return self.data ^ other
    __rxor__ = __xor__

    def __lshift__(self, other):
        return self.data << other

    def __rlshift__(self, other):
        return other << self.data

    def __rshift__(self, other):
        return self.data >> other

    def __rrshift__(self, other):
        return other >> self.data

    def __invert__(self):
        return ~self.data

    def __nonzero__(self):
        return bool(self.data)


class FileResponse(BaseResponse):
    @property
    def json(self):
        raise NotImplementedError

    @property
    def data(self):
        return self.response.content

    @property
    def filename(self):
        disposition = self.response.headers.get('content-disposition')
        return get_filename(disposition)


def get_default_request_handler():
    """Return the default request handler."""
    if not _DEFAULT_REQUEST_HANDLER:
        raise Exception("Request handler required but no default was found.")
    return _DEFAULT_REQUEST_HANDLER


def set_default_request_handler(request_handler):
    """Set the default request handler."""
    if not isinstance(request_handler, RequestHandler):
        raise TypeError(
            "Attempted to set an invalid request handler as default.")
    global _DEFAULT_REQUEST_HANDLER
    _DEFAULT_REQUEST_HANDLER = request_handler



from six.moves import builtins
import io


# taken from functools.total_ordering
def total_ordering(cls):  # pragma: no cover
    """Class decorator that fills in missing ordering methods"""
    convert = {
        '__lt__': [('__gt__', lambda self, other: not (self < other or self == other)),
                   ('__le__', lambda self, other: self < other or self == other),
                   ('__ge__', lambda self, other: not self < other)],
        '__le__': [('__ge__', lambda self, other: not self <= other or self == other),
                   ('__lt__', lambda self, other: self <= other and not self == other),
                   ('__gt__', lambda self, other: not self <= other)],
        '__gt__': [('__lt__', lambda self, other: not (self > other or self == other)),
                   ('__ge__', lambda self, other: self > other or self == other),
                   ('__le__', lambda self, other: not self > other)],
        '__ge__': [('__le__', lambda self, other: (not self >= other) or self == other),
                   ('__gt__', lambda self, other: self >= other and not self == other),
                   ('__lt__', lambda self, other: not self >= other)]
    }
    roots = set(dir(cls)) & set(convert)
    if not roots:
        raise ValueError('must define at least one ordering operation: < > <= >=')
    root = max(roots)       # prefer __lt__ to __le__ to __gt__ to __ge__
    for opname, opfunc in convert[root]:
        if opname not in roots:
            opfunc.__name__ = opname
            opfunc.__doc__ = getattr(int, opname).__doc__
            setattr(cls, opname, opfunc)
    return cls


file_type = file if hasattr(builtins, 'file') else io.IOBase



import copy

from six import string_types

from ubersmith.api import (
    METHODS,
    BaseResponse,
    DictResponse,
    IntResponse,
    FileResponse,
    get_default_request_handler,
)
from ubersmith.exceptions import ValidationError

__all__ = [
    # abstract call classes
    'BaseCall',
    # generate generic calls
    'generate_generic_calls',
    # concrete call classes
    'client',
    'device',
    'order',
    'sales',
    'support',
    'uber',
]


class BaseCall(object):
    """Abstract class to implement a call with validation/cleaning/etc."""

    method = ''  # ubersmith method name, should be defined on child classes
    required_fields = []  # field names that should be present in request_data
    cleaner = None  # function to clean response (see ubersmith.clean)

    def __init__(self, request_data=None, request_handler=None):
        """Setup call with provided request data and handler."""
        self.request_data = request_data or {}  # data for the request
        self.request_handler = request_handler or \
            get_default_request_handler()  # handler to fullfil the request
        self.response = None  # response is stored here

    def render(self):
        """Validate, process, clean and return the result of the call."""
        if not self.validate():
            raise ValidationError

        self.process_request()
        self.clean()

        return self.response

    def validate(self):
        """Validate request data before sending it out. Return True/False."""
        # check if required_fields aren't present
        for field in set(self.required_fields) - set(self.request_data):
            if not isinstance(field, string_types):
                # field was a collection, iterate over it and check by OR
                return bool(set(field) & set(self.request_data))
            return False
        return True

    def process_request(self):
        """Processing the call and set response_data."""
        self.response = self.request_handler.process_request(
            self.method, self.request_data)

    def clean(self):
        """Clean response."""
        if self.response.type == 'application/json':
            cleaned = copy.deepcopy(self.response.data)
            if self.cleaner is not None:
                cleaned = self.cleaner(cleaned)

            typed_response = {
                dict: DictResponse,
                int: IntResponse,
            }.get(type(cleaned), BaseResponse)
            self.response = typed_response.from_cleaned(self.response, cleaned)
        else:
            self.response = FileResponse(self.response.response)


def _get_call_class(method):
    """Find the call class for method if it exists else create one."""
    call_base, call_name = method.split('.', 1)
    # import the call class's module
    mod = __import__('ubersmith.calls.{0}'.format(call_base), fromlist=[''])
    # grab all the public members of the module
    gen = (getattr(mod, x) for x in dir(mod) if not x.startswith('_'))
    # filter them down to subclasses of BaseCall
    gen = (x for x in gen if type(x) is type and issubclass(x, BaseCall))
    # return first one that matches our method
    for call_class in gen:
        if call_class.method == method:
            return call_class
    else:
        class GenericCall(BaseCall):
            method = '.'.join((call_base, call_name))
        return GenericCall


class GenericCall(object):
    def __init__(self, call_class, request_handler=None):
        self.call_class = call_class
        self.request_handler = request_handler

    def handler(self, request_handler):
        """Return new GenericCall that is bound to a request handler."""
        return GenericCall(self.call_class, request_handler)

    def __call__(self, **kwargs):
        return self.call_class(kwargs, self.request_handler).render()


def generate_generic_calls(base, ns):
    # get all valid methods with base
    methods = (m for m in METHODS if m.split('.', 1)[0] == base)
    for method in methods:
        call_name = method.split('.', 1)[1]
        if call_name not in ns:
            # find the appropriate class
            call_class = _get_call_class(method)
            # create a call function and stick it in the namespace
            generic_call = GenericCall(call_class)
            generic_call.__name__ = str(call_name)
            generic_call.__doc__ = METHODS[method]
            # this may or may not be a good idea, see:
            # http://stackoverflow.com/questions/10113892/semantics-of-module
            # generic_call.__module__ = 'ubersmith.{0}'.format(base)
            ns[call_name] = generic_call
            # add call to __all__ if needed
            if '__all__' in ns and call_name not in ns['__all__']:
                ns['__all__'].append(call_name)
        else:
            if not ns[call_name].__doc__:
                ns[call_name].__doc__ = METHODS[method]



from ubersmith.api import RequestHandler, set_default_request_handler

__all__ = [
    # package modules
    'api',
    'exceptions',
    'utils',
    # call classes
    'calls',
    # call functions
    'client',
    'device',
    'order',
    'sales',
    'support',
    'uber',
    # init function
    'init',
]


def init(base_url, username=None, password=None, verify=True):
    """Initialize ubersmith API module with HTTP request handler."""
    handler = RequestHandler(base_url, username, password, verify)
    set_default_request_handler(handler)
    return handler



"""Exceptions for ubersmith library."""

__all__ = [
    'UbersmithError',
    'RequestError',
    'ValidationError',
    'ResponseError',
    'UpdatingTokenResponse',
    'MaintenanceResponse',
]


class UbersmithError(Exception):
    msg = "An error occured with the Ubersmith API."

    def __init__(self, msg=None):
        super(UbersmithError, self).__init__()
        if msg is not None:
            self.msg = msg

    def __str__(self):
        return self.msg


class RequestError(UbersmithError):
    msg = "An error occured with the request to the Ubersmith API."


class ValidationError(RequestError):
    msg = "The request data was invalid for the Ubersmith request."


class ResponseError(UbersmithError):
    """Exception for Ubersmith API Response.

        msg: optional message to pass along w/ stacktrace
        response: decoded json response w/ error info

    """
    msg = "An error occured in the response from the Ubersmith API."

    def __init__(self, msg=None, response=None):
        super(ResponseError, self).__init__(msg)
        if response is not None:
            self.response = response
            self.error_code = response.get('error_code')
            self.error_message = response.get('error_message')
            self.msg = '{0.msg} {0.error_code} {0.error_message}'.format(self)


class UpdatingTokenResponse(ResponseError):
    msg = "Ubersmith is updating token."


class MaintenanceResponse(ResponseError):
    msg = "Ubersmith is currently undergoing maintenance."



import json

from six import text_type

import ubersmith
import ubersmith.api
import ubersmith.uber


def setup_module():
    ubersmith.init(**{
        'base_url': '',
        'username': '',
        'password': '',
    })


def teardown_module():
    ubersmith.api._DEFAULT_REQUEST_HANDLER = None


def test_responses_can_getitem(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {
            "foo": "bar",
            "nested": {
                "baz": "qux",
            }
        },
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    result = ubersmith.uber.method_list()
    assert result["foo"] == "bar"
    assert result["nested"]["baz"] == "qux"


def test_responses_can_be_turned_into_dicts(response):
    resp_json = {
        "status": True,
        "error_code": None,
        "error_message": "",
        "data": {
            "foo": "bar",
        },
    }
    response.json.return_value = resp_json
    response.content = json.dumps(resp_json)
    response.text = text_type(response.content)
    expected = {
        "foo": "bar",
    }
    assert dict(ubersmith.uber.method_list()) == expected



from setuptools import setup, find_packages


setup(
    name='ubersmith',
    version='0.3.6',
    author='Jason Keene',
    author_email='jasonkeene@gmail.com',
    description='Client library for the Ubersmith API 2.0',
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=['tests']),
    install_requires=open('requirements.txt').read().strip().split('\n'),
    url='https://github.com/jasonkeene/python-ubersmith',
    license='MIT License',
    keywords=['ubersmith'],
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import datetime
import six
import time
from decimal import Decimal
try:
    import __builtin__ as builtins
except ImportError:  # pragma: no cover
    import builtins

import phpserialize


_CLEANERS = {}


def cleaner(func):
    _CLEANERS[func.__name__] = func
    return func


@cleaner
def php(val):
    return phpserialize.loads(val.encode("utf-8"))


@cleaner
def timestamp(val):
    return datetime.datetime.fromtimestamp(float(val))


@cleaner
def date(val):
    return datetime.date(*time.strptime(val, '%b/%d/%Y')[:3])


@cleaner
def decimal(val):
    if val is None:
        return None
    if isinstance(val, six.string_types):
        val = val.replace(',', '')
        if val.strip() == '':
            return None
    return Decimal(val)


@cleaner
def int(val):
    if val is None:
        return None
    if isinstance(val, six.string_types):
        val = val.replace(',', '')
        if val.strip() == '':
            return None
    return builtins.int(val)


class clean(object):
    def __init__(self, cleaner, keys=None, values=None, raises=False):
        self.cleaner = cleaner if callable(cleaner) else _CLEANERS[cleaner]
        if keys is not None and not callable(keys):
            if type(keys) is dict:
                for k, v in keys.items():
                    keys[k] = v if callable(v) else _CLEANERS[v]
            else:
                keys = _CLEANERS[keys]
        self.keys = keys
        # TODO: write tighter tests to cover these branches
        if values is not None and not callable(values):
            if type(values) is dict:
                for k, v in values.items():
                    values[k] = v if callable(v) else _CLEANERS[v]
            else:
                values = _CLEANERS[values]
        self.values = values
        self.raises = raises

    def __call__(self, val):
        val = self.cleaner(val)
        if self.cleaner is list:
            val = self._clean_list(val)
        elif self.cleaner is dict:
            val = self._clean_dict(val)
        return val

    def _clean_list(self, val):
        # TODO: factor keys/values cleaning out to methods
        # clean values
        if self.values is not None:
            if callable(self.values):
                # apply cleaner to all elements
                cleaner = self.values
                for i, element in enumerate(val):
                    val[i] = cleaner(element)
            else:
                # apply cleaners to specific values
                for i, cleaner in self.values.items():
                    try:
                        val[i] = cleaner(val[i])
                    except IndexError as e:
                        if self.raises:
                            raise e
        return val

    def _clean_dict(self, val):
        # TODO: factor keys/values cleaning out to methods
        # clean keys
        if self.keys is not None:
            tmp = {}
            if callable(self.keys):
                # apply cleaner to all keys
                cleaner = self.keys
                for k, v in val.items():
                    tmp[cleaner(k)] = v
            else:
                # apply cleaners to specific keys
                for k, v in val.items():
                    cleaner = self.keys.get(k, lambda x: x)
                    tmp[cleaner(k)] = val[k]
            val = tmp

        # TODO: factor keys/values cleaning out to methods
        # clean values
        if self.values is not None:
            if callable(self.values):
                # apply cleaner to all elements
                cleaner = self.values
                for k, v in val.items():
                    val[k] = cleaner(val[k])
            else:
                # apply cleaners to specific values
                for k, cleaner in self.values.items():
                    try:
                        val[k] = cleaner(val[k])
                    except KeyError as e:
                        if self.raises:
                            raise e
        return val



import pytest

import ubersmith
import ubersmith.api
import ubersmith.uber


def setup_module():
    ubersmith.init(**{
        'base_url': '',
        'username': '',
        'password': '',
    })


def teardown_module():
    ubersmith.api._DEFAULT_REQUEST_HANDLER = None


def test_uber_documentation(response):
    response.headers = {
        'content-type': 'application/pdf',
        'content-disposition': 'inline; filename="doc.pdf";'
    }
    response.text = u"Some PDF data."
    response.content = str(response.text)
    uberfile = ubersmith.uber.documentation()
    assert uberfile.type == 'application/pdf'
    assert uberfile.filename == "doc.pdf"
    assert str(uberfile.data) == response.text



"""Sales call classes.

These classes implement any response cleaning and validation needed.  If a
call class isn't defined for a given method then one is created using
ubersmith.calls.BaseCall.

"""

# from ubersmith.calls import BaseCall, GroupCall
from ubersmith.utils import prepend_base

__all__ = []

_ = prepend_base(__name__.split('.')[-1])


# class (BaseCall):
#     method = _('')
